open Gui

(* var_t *)
module VarS = struct
 type 'var_t t = V of string

  let map f v = match v with V (s) -> V (s)

  let unify f v1 v2 = match (v1, v2) with
      (V (str1), V (str2)) when str1 = str2 -> ()
    | _ -> raise (Unify_Error "of VarS")

  let draw v = match v with V (s) -> create_str s

  let search _ = false
end

module Var  = Fix.F (VarS)

let var n = Var.make_fix (VarS.V (n))

(* tvar_t *)
module TVarS = struct
  type 'tvar_t t = TV of string

  let map f v = match v with TV (s) -> TV (s)

  let unify f v1 v2 = match (v1, v2) with
      (TV (str1), TV (str2)) when str1 = str2 -> ()
    | _ -> raise (Unify_Error "of TVarS")

  let draw v = match v with TV (s) -> create_str s

  let search _ = false
end

module TVar = Fix.F (TVarS)

let tv x = TVar.make_fix (TVarS.TV (x))

(* type_t *)
module TypeS = struct
  type 'type_t t = TInt
		 | TVar of TVar.fix_t
		 | Fun of 'type_t * 'type_t

  let map f tp = match tp with
      TInt -> TInt
    | TVar (tv) -> TVar (tv)
    | Fun (tp1, tp2) -> Fun (f tp1, f tp2)

  let unify f tp1 tp2 = match (tp1, tp2) with
      (TInt, TInt) -> ()
    | (TVar (tv1), TVar (tv2)) -> TVar.unify tv1 tv2
    | (Fun (tp1, tp2), Fun (tp3, tp4)) -> f tp1 tp3; f tp2 tp4
    | _ -> raise (Unify_Error "of Type")

  let search tp = match tp with
      TInt -> false
    | TVar (tv) -> TVar.search tv
    | Fun (tp1, tp2) -> tp1 || tp2

  let draw tp = match tp with
      TInt -> create_str "int"
    | TVar (tv) -> TVar.draw tv
    | Fun (tp1, tp2) ->
	combineH[create_str "("; tp1; create_str "→"; tp2; create_str ")"]
	(* tp1 が arrow なら "(tp1)" にしたい *)
end

module Type = Fix.F (TypeS)

let tint = Type.make_fix (TypeS.TInt)
let tvar tv = Type.make_fix (TypeS.TVar (tv))
let tfun tp1 tp2 = Type.make_fix (TypeS.Fun (tp1, tp2))

(* ts_t *)
module TsS = struct
  type 'ts_t t = Forall of TVar.fix_t list * Type.fix_t

  let map f ts = match ts with Forall(tv_lst, tp) -> Forall(tv_lst, tp)

  let search tss = match tss with
      Forall(tv_lst, tp) ->
	(* b1 の計算、直した方が良い？ *)
	let b1 = List.fold_left (fun b tv -> TVar.search tv || b) false tv_lst in
	let b2 = Type.search tp in
	b1 || b2

  let unify f ts1 ts2 = match (ts1, ts2) with
      (Forall(tv_lst1, tp1), Forall(tv_lst2, tp2)) ->
	List.iter2 TVar.unify tv_lst1 tv_lst2; Type.unify tp1 tp2

  let draw ts = match ts with
      Forall(tv_lst, tp) ->
	if tv_lst = []
	then Type.draw tp
	else
	  let id_lst = List.map TVar.draw tv_lst in
	  let tp_id = Type.draw tp in
	  combineH (create_str "∀" :: id_lst @ [create_str "."; tp_id])
end

module Ts = Fix.F (TsS)

let forall lst tp = Ts.make_fix (TsS.Forall (lst, tp))

(* term_t *)
module TermS = struct
  type 'term_t t = Var of Var.fix_t
		 | Abs of Var.fix_t * Type.fix_t * 'term_t
		 | App of 'term_t * 'term_t
		 | Let of Var.fix_t * 'term_t * 'term_t

  let map f tm = match tm with
      Var (v) -> Var (v)
    | Abs (v, tp, f1) -> Abs (v, tp, f f1)
    | App (f1, f2) -> App (f f1, f f2)
    | Let (v, f1, f2) -> Let (v, f f1, f f2)

  let search tm = match tm with
      Var (v) -> Var.search v
    | Abs (v, tp, tm) ->
	let b1 = Var.search v in
	let b2 = Type.search tp in
	b1 || b2 || tm
    | App (tm1, tm2) -> tm1 || tm2
    | Let (v, tm1, tm2) -> Var.search v || tm1 ||tm2

  let unify f tm1 tm2 =  match (tm1, tm2) with
      (Var (v1), Var (v2)) -> Var.unify v1 v2
    | (Abs (v1, tp1, f1), Abs (v2, tp2, f2)) ->
	Var.unify v1 v2;
	Type.unify tp1 tp2;
	f f1 f2
    | (App (f1, f2), App (f3, f4)) -> f f1 f3; f f2 f4
    | (Let (v1, f1, f3), Let (v2, f2, f4)) ->
	Var.unify v1 v2; f f1 f2; f f3 f4
    | _ -> raise (Unify_Error "of Term")

  let draw tm = match tm with
      Var (v) -> Var.draw v
    | Abs (v, tp, f) ->
	combineH [create_str "λ"; Var.draw v; create_str ":";
		  Type.draw tp; create_str ". "; f]
    | App(fix1, fix2) ->
	combineH [create_str "("; fix1; create_str " "; fix2; create_str ")"]
    | Let(v, f1, f2) ->
	combineH[create_str "let "; Var.draw v; create_str "="; f1;
		 create_str " in "; f2]
end

module Term = Fix.F (TermS)

let term_var v = Term.make_fix (TermS.Var (v))
let term_abs x t e = Term.make_fix (TermS.Abs (x, t, e))
let term_app e1 e2 = Term.make_fix (TermS.App (e1, e2))
let term_let x tm1 tm2 = Term.make_fix (TermS.Let (x, tm1, tm2))

(* env_t *)
module EnvS = struct
  type elem_t = E of Var.fix_t * Ts.fix_t

  type 'env_t t = Empty
		| Cons of elem_t * 'env_t

  let map f env = match env with
      Empty -> Empty
    | Cons (e, env) -> Cons (e, f env)

  let search_e elm = match elm with
      E (v, ts) ->
	let b1 = Var.search v in
	let b2 = Ts.search ts in
	b1 || b2

  let search env = match env with
      Empty -> false
    | Cons (elm, env) -> search_e elm || env

  let unify_e e1 e2 = match (e1, e2) with
      (E (v1, ts1), E (v2, ts2)) ->
	Var.unify v1 v2; Ts.unify ts1 ts2

  let unify f env1 env2 = match (env1, env2) with
      (Empty, Empty) -> ()
    | (Cons (e1, env1), Cons (e2, env2)) ->
	unify_e e1 e2;
	f env1 env2
    | _ -> raise (Unify_Error "of Env")

  let draw_e e = match e with
      E (v, ts) -> combineH [Var.draw v; create_str ":"; Ts.draw ts]

  let draw env = match env with
      Empty -> create_str "Φ"
    | Cons (e, env) ->
	combineH[draw_e e; create_str ", "; env]
end

module Env = Fix.F (EnvS)

let empty = Env.make_fix (EnvS.Empty)
let cons x ts e = Env.make_fix (EnvS.Cons (EnvS.E (x, ts), e))

(* judge_t *)
module JudgeS = struct
  type 'judge_t t = Judge of Env.fix_t * Term.fix_t * Type.fix_t
		  | Equal of Type.fix_t * Type.fix_t
		  | EqualTs of Ts.fix_t * Ts.fix_t

  let map f j = match j with
      Judge (e, tm, tp) -> Judge (e, tm, tp)
    | Equal (tp1, tp2) -> Equal (tp1, tp2)
    | EqualTs (tp1, tp2) -> EqualTs (tp1, tp2)

  let search j = match j with
      Judge (env, tm, tp) ->
	let b1 = Env.search env in
	let b2 = Term.search tm in
	let b3 = Type.search tp in
	b1 || b2 || b3
    | Equal (tp1, tp2) ->
	let b1 = Type.search tp1 in
	let b2 = Type.search tp2 in
	b1 || b2
    | EqualTs (ts1, ts2) ->
	let b1 = Ts.search ts1 in
	let b2 = Ts.search ts2 in
	b1 || b2

  let unify f j1 j2 = match (j1, j2) with
      (Judge (env1,tm1,tp1), Judge (env2,tm2,tp2)) ->
	Env.unify env1 env2;
	Term.unify tm1 tm2;
	Type.unify tp1 tp2
    | (Equal (tp1,tp2), Equal (tp3,tp4)) ->
	Type.unify tp1 tp3;
	Type.unify tp2 tp4
    | (EqualTs (tp1,tp2), EqualTs (tp3,tp4)) ->
	Ts.unify tp1 tp3;
	Ts.unify tp2 tp4
    | _ -> raise (Unify_Error "of Judge")

  let draw j = match j with
      Judge(env, tm, tp) ->
	 combineH[Env.draw env; create_str "├ "; Term.draw tm;
		  create_str " : "; Type.draw tp]
    | Equal(tp1, tp2) ->
	combineH[Type.draw tp1; create_str " = "; Type.draw tp2]
    | EqualTs(ts1, ts2) ->
	combineH[Ts.draw ts1; create_str " = "; Ts.draw ts2]
end

module Judge = Fix.F(JudgeS)

let judge e tm tp = Judge.make_fix (JudgeS.Judge (e, tm, tp))
let equal tp1 tp2 = Judge.make_fix (JudgeS.Equal (tp1, tp2))
let equalts ts1 ts2 = Judge.make_fix (JudgeS.EqualTs (ts1, ts2))

(* ユーザ定義の関数 *)
(* いつも fix を受け取ってくるように書く *)

let rec fold_type tint tvar tfun typ = match Type.take typ with
    TypeS.TInt -> tint
  | TypeS.TVar (tv) -> tvar tv
  | TypeS.Fun (typ1, typ2) ->
      tfun (fold_type tint tvar tfun typ1)
	   (fold_type tint tvar tfun typ2)

let replace tp alist =
  fold_type tint
	    (fun tv -> (try
			  List.assoc (TVar.take tv) alist
			with Not_found -> tvar tv)) tfun tp

(* 環境から型を取ってきて、それの instance を返す *)
let rec instance ts = match Ts.take ts with
  TsS.Forall(tv_lst, tp) ->
    let alist = List.map (fun tv -> (TVar.take tv, Type.make_meta("i")))
			 tv_lst in
    if alist = [] then tp else
    replace tp alist

(* 環境から変数に当たる型を取ってくる *)
let rec get env var =
  let rec loop env = match Env.take env with
      EnvS.Empty ->
	Type.make_code
	  (fun () -> combineH [create_str "get(";
			       Env.draw env;
			       create_str ",";
			       Var.draw var;
			       create_str ")"])
	  (fun () -> get env var)
    | EnvS.Cons (EnvS.E(var2, ts), rest) ->
	if Var.take var = Var.take var2
	then instance ts
	else loop rest
  in try
      loop env
     with Cannot_take -> Type.make_code
       (fun () -> combineH [create_str "get(";
			    Env.draw env;
			    create_str ",";
			    Var.draw var;
			    create_str ")"])
       (fun () -> get env var)

(* 推論規則を定義する *)

(* TVAR *)
let t_var () =
  let x = Var.make_meta "x" in
  let t = Type.make_meta "tp" in
  let e = Env.make_meta "e" in
  Infer(judge e (term_var x) t,
	[Infer(equal t (get e x),
	       [])])

(* TABS *)
let t_abs () =
  let e = Env.make_meta "e" in
  let x = Var.make_meta "x" in
  let tm = Term.make_meta "t" in
  let tp1 = Type.make_meta "T" in
  let ts1 = forall [] tp1 in
  let tp2 = Type.make_meta "T" in
  Infer(judge e (term_abs x tp1 tm) (tfun tp1 tp2),
	[Infer(judge (cons x ts1 e) tm tp2,
	       [])])

(* TAPP *)
let t_app () =
  let e = Env.make_meta "e" in
  let tm1 = Term.make_meta "t" in
  let tm2 = Term.make_meta "t" in
  let tp1 = Type.make_meta "T" in
  let tp2 = Type.make_meta "T" in
  Infer(judge e (term_app tm1 tm2) tp2,
	[Infer(judge e tm1 (tfun tp1 tp2),
	       []);
	 Infer(judge e tm2 tp1,
	       [])])

let rec append lst1 lst2 = match lst1 with
    [] -> lst2
  | first :: rest -> if List.mem first lst2 then append rest lst2
					    else first :: append rest lst2

let rec minus lst1 lst2 = match lst1 with
    [] -> []
  | first :: rest -> if List.mem first lst2 then append rest lst2
					    else first :: append rest lst2

let free_tvars tp = fold_type [] (fun tv -> [tv]) append tp

let free_tvars_ts ts = match Ts.take ts with
    TsS.Forall(tv_lst, tp) -> minus (free_tvars tp) tv_lst

let rec free_tvars_e env = match Env.take env with
    EnvS.Empty -> []
  | EnvS.Cons (EnvS.E(_, ts), rest) ->
      append (free_tvars_ts ts) (free_tvars_e rest)

let rec gen e tp = try
  let tvars = free_tvars tp in
  let tvars_e = free_tvars_e e in
  let tvars' = minus tvars tvars_e in
  forall tvars' tp
with Gui.Cannot_take ->
  Ts.make_code
    (fun () -> combineH [create_str "gen ("; Env.draw e; create_str ";";
			 Type.draw tp; create_str ")"])
    (fun () -> gen e tp)

(* TLET *)
let t_let () =
  let e = Env.make_meta "e" in
  let x = Var.make_meta "x" in
  let tm1 = Term.make_meta "t" in
  let tm2 = Term.make_meta "t" in
  let tp1 = Type.make_meta "T" in
  let tp2 = Type.make_meta "T" in
  let ts = Ts.make_meta "S" in
  Infer(judge e (term_let x tm1 tm2) tp2,
	[Infer(judge e tm1 tp1, []);
	 Infer(equalts ts (gen e tp1), []);
	 Infer(judge (cons x ts e) tm2 tp2, [])])

(* TEQUAL *)
let t_equal () =
  let tp = Type.make_meta "T" in
  Axiom(equal tp tp)

(* TEQUALTS *)
let t_equalts () =
  let ts = Ts.make_meta "T" in
  Axiom(equalts ts ts)

(* 登録される推論規則 *)
let infer_rule_list =
  [("TVAR", t_var); ("TABS", t_abs); ("TAPP", t_app);
   ("TLET", t_let); ("TEQUAL", t_equal); ("TEQUALTS", t_equalts)]
